/*
 * Nick Cross
 * CSC 570R Final Project
 * Particle system using Nvidia's CG
 *
 *------------------------------------------------------------------*/


#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>



#include <cg\cg.h>														// NEW: Cg Header
#include <cg\cggl.h>													// NEW: Cg OpenGL Specific Header
#include <gl\gl.h>														// Header File For The OpenGL32 Library
#include <gl\glu.h>
#include  <glut.h>

using namespace std;

#pragma comment( lib, "opengl32.lib" )									// Search For OpenGL32.lib While Linking
#pragma comment( lib, "glu32.lib" )										// Search For GLu32.lib While Linking

#pragma comment( lib, "cg.lib" )										// Search For Cg.lib While Linking
#pragma comment( lib, "cggl.lib" )										// Search For CgGL.lib While Linking

// some function prototypes
void display(void);
void sumForces(void);
void EulerIntegrate(void);
void normalize(float[3]);
void normCrossProd(float[3], float[3], float[3]);
void  initializeParticleDataStructure(int);


#define MAX_X_DIST 10
#define MAX_Z_DIST 10
#define PEOPLE_PREC 2000
#define MAX_Z_VEL 5
#define MAX_X_VEL 3
#define Z_PREC 1750
#define X_PREC 2250
#define MIN_Y_VEL 6
#define MAX_Y_VEL 15
#define Y_PREC 1500
#define WAIT_TIME 5
#define MAX_CAM_VERT 1.5
#define MIN_CAM_VERT -1.5
#define PI 3.14159265
#define MAX_MASS 2


// no provision to move viewpoint in this code,
//   but here are the transforms anyway
// initial viewer position
static GLdouble viewer[] = {0.0, 0.0, 10.0};
// initial viewer angle
static GLfloat theta[] = {0.0, 0.0, 0.0};

// animation variables
static int frame = 0;
static int startFrame = 0;
static int endFrame = 100;
static int increment = 1;

const int NUM_PARTICLES = 10000;
const float GRAVITY = 9.8;
const float DRAG = 0.1;
const float TIMESTEP = 0.1; 

float mass[NUM_PARTICLES];
float pos[NUM_PARTICLES][3];
float vel[NUM_PARTICLES][3];
float acc[NUM_PARTICLES][3];
float force[NUM_PARTICLES][3];
int on2D;

float timeCnt[NUM_PARTICLES];
int firstTime = 1;

float count;
float myCnt;

static GLuint textMap = 0;




struct _CGcontext {

};

struct _CGprogram {

};

struct _CGparameter {

};

/*data structure for the image used for  texture mapping */
typedef struct Image {
  unsigned long sizeX;
  unsigned long sizeY;
  char *data;
} Image;

typedef struct RGB {
  GLubyte r;
  GLubyte g;
  GLubyte b;
} RGB;

typedef struct Vector3{
  float x;
  float y;
  float z;
} Vector3;
Vector3 eye;
Vector3 eyeTarget;
float phi;
float mytheta;
float radius;
Vector3 eyeStart;

Image *TextureImage;
RGB myimage[16][16];
RGB* pixel;

int ImageLoad(char*, Image*); 
int GW;
int GH;

int startx, starty;
CGcontext context;
CGprogram	program, fragProgram;													// Our Cg Vertex Program
CGprofile	vertexProfile, fragProfile;											// The Profile To Use For Our Vertex Shader
CGparameter	pInitial, vInitial, tInitial, modelViewMatrix, cgForce, gravity, globalTime, modelViewProj, cgMass;//acc,vel;						// The Parameters Needed For Our Shader

// Pixel to world x-coor.  Translates to image first, then world.
float p2w_x(int pt){
	float r = 2 * (float)GW / (float)GH;
	float l = -2 * ((float)GW / (float)GH);
	float iX = ((float)(2 * pt - GW + 1) / (float)GW);
	float wX = ((float)(iX * (r - l) + r + l)) / (float)2;	

	return wX;
}

// Pixel to image x-coor
float p2i_x(int pt){
	float r = 2 * (float)GW / (float)GH;
	float l = -2 * ((float)GW / (float)GH);
	float iX = ((float)(2 * pt - GW + 1) / (float)GW);
	return iX;
}

// Pixel to image y-coor
float p2i_y(int pt){
	float newY = GH - pt - 1;
	float t = 2.0;
	float b = -2.0;
	float iY = ((float)(2 * newY - GH + 1) / (float)GH);
	return iY;
}

// Pixel to world y-coor.  Translates to image first, then world.
float p2w_y(int pt){
	float newY = GH - pt - 1;
	float t = 2.0;
	float b = -2.0;
	float iY = ((float)(2 * newY - GH + 1) / (float)GH);
	float wY = (float)((float)((t - b) * iY + t + b) / (float)2);	
	return wY;
}

static unsigned int getshort(FILE *fp){
  int c, c1;

  /* get 2 bytes*/
  c = getc(fp);
  c1 = getc(fp);

  return ((unsigned int) c) + (((unsigned int) c1) << 8);
}

/*
* getint and getshort are help functions to load the bitmap byte by byte
*/
static unsigned int getint(FILE *fp) {
  int c, c1, c2, c3;

  /*  get 4 bytes */
  c = getc(fp);
  c1 = getc(fp);
  c2 = getc(fp);
  c3 = getc(fp);

  return ((unsigned int) c) +
    (((unsigned int) c1) << 8) +
    (((unsigned int) c2) << 16) +
    (((unsigned int) c3) << 24);
}

//initialization for texture mapping
void init_tex() {
  glBindTexture(GL_TEXTURE_2D, 0);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, 64, 64, 0, GL_RGB, GL_UNSIGNED_BYTE, pixel);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

/*  quick and dirty bitmap loader...for 24 bit bitmaps with 1 plane only.  */

//routines to load in a bmp files - must be 2^nx2^m and a 24bit bmp
GLvoid LoadTexture(char* image_file, int texID) {
 
  TextureImage = (Image *) malloc(sizeof(Image));
  if (TextureImage == NULL) {
    printf("Error allocating space for image");
    exit(1);
  }
  cout << "trying to load " << image_file << endl;
  if (!ImageLoad(image_file, TextureImage)) {
    exit(1);
  }
  /*  2d texture, level of detail 0 (normal), 3 components (red, green, blue),            */
  /*  x size from image, y size from image,                                              */
  /*  border 0 (normal), rgb color data, unsigned byte data, data  */
  glBindTexture(GL_TEXTURE_2D, texID);
  glTexImage2D(GL_TEXTURE_2D, 0, 3,
    TextureImage->sizeX, TextureImage->sizeY,
    0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage->data);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST); /*  cheap scaling when image bigger than texture */
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); /*  cheap scaling when image smalled than texture*/
 
}

int ImageLoad(char *filename, Image *image) {
  FILE *file;
  unsigned long size;                 /*  size of the image in bytes. */
  unsigned long i;                    /*  standard counter. */
  unsigned short int planes;          /*  number of planes in image (must be 1)  */
  unsigned short int bpp;             /*  number of bits per pixel (must be 24) */
  char temp;                          /*  used to convert bgr to rgb color. */

  /*  make sure the file is there. */
  if ((file = fopen(filename, "rb"))==NULL) {
    printf("File Not Found : %s\n",filename);
    return 0;
  }

  /*  seek through the bmp header, up to the width height: */
  fseek(file, 18, SEEK_CUR);

  /*  No 100% errorchecking anymore!!! */

  /*  read the width */    image->sizeX = getint (file);

  /*  read the height */
  image->sizeY = getint (file);

  /*  calculate the size (assuming 24 bits or 3 bytes per pixel). */
  size = image->sizeX * image->sizeY * 3;

  /*  read the planes */
  planes = getshort(file);
  if (planes != 1) {
    printf("Planes from %s is not 1: %u\n", filename, planes);
    return 0;
  }

  /*  read the bpp */
  bpp = getshort(file);
  if (bpp != 24) {
    printf("Bpp from %s is not 24: %u\n", filename, bpp);
    return 0;
  }
/*  seek past the rest of the bitmap header. */
  fseek(file, 24, SEEK_CUR);

  /*  read the data.  */
  image->data = (char *) malloc(size);
  if (image->data == NULL) {
    printf("Error allocating memory for color-corrected image data");
    return 0;
  }

  if ((i = fread(image->data, size, 1, file)) != 1) {
    printf("Error reading image data from %s.\n", filename);
    return 0;
  }

  for (i=0;i<size;i+=3) { /*  reverse all of the colors. (bgr -> rgb) */
    temp = image->data[i];
    image->data[i] = image->data[i+2];
    image->data[i+2] = temp;
  }

  fclose(file); /* Close the file and release the filedes */

  /*  we're done. */
  return 1;
}

void keyCallback(unsigned char key, int x, int y) {
   // move viewer with x, y, and z keys
   // capital moves in + direction, lower-case - direction

   if(key == 'd') {force[0][0] += .1; cgGLSetParameter3f(cgForce, force[0][0], force[0][1], force[0][2]);}
   if(key == 'a') {force[0][0] -= .1; cgGLSetParameter3f(cgForce, force[0][0], force[0][1], force[0][2]);}
   if(key == 'w') {force[0][1] += .1; cgGLSetParameter3f(cgForce, force[0][0], force[0][1], force[0][2]);}
   if(key == 's') {force[0][1] -= .1; cgGLSetParameter3f(cgForce, force[0][0], force[0][1], force[0][2]);}
   if(key == 'q') {force[0][2] += .1; cgGLSetParameter3f(cgForce, force[0][0], force[0][1], force[0][2]);}
   if(key == 'e') {force[0][2] -= .1; cgGLSetParameter3f(cgForce, force[0][0], force[0][1], force[0][2]);}
   if(key == 'm') {on2D = !on2D; printf("On2d turned off\n");}

   display();
}

// The main mouse callback function
void mouse(int button, int state, int x, int y) {
  
  if (button == GLUT_LEFT_BUTTON) {
    if (state == GLUT_DOWN) { /* if the left button is clicked */
      
      startx = x;
      starty = GH-y-1;	  
	  
		  // Get initial spots
		  eyeStart.x = p2w_x(x);
		  eyeStart.y = p2w_y(y);
		  
	  
    }

	// On mouse release
	else if (state == GLUT_UP){
		
			
		
		
	}
  }
  glutPostRedisplay();
}

// Main mouse movement callback
void mouseMove(int x, int y) {  

  
 
	  // Temp vector so I dont have to calculate world coords often
	  Vector3 temp;
	  temp.x = p2w_x(x);
	  temp.y = p2w_y(y);
	  //temp.z = getZ(temp.x, temp.y);

	  // Increment theta and phi
	  mytheta += eyeStart.x - temp.x;
	  phi += eyeStart.y - temp.y;

	  // Check vertical limitations
	  if(phi > MAX_CAM_VERT){
		  phi = MAX_CAM_VERT;
	  }
	  else if(phi < MIN_CAM_VERT){
		  phi = MIN_CAM_VERT;
	  }

	  // Update the camera
	  eye.x = eyeTarget.x + radius * cos(-phi) * cos(1.57 - mytheta);
	  eye.y = eyeTarget.y + radius * sin(-phi);
	  eye.z = eyeTarget.z + radius * cos(-phi) * cos(mytheta);
		
	  
	  

	  // Update for last step
	  eyeStart.x = temp.x;
	  eyeStart.y = temp.y;
	  

     display();
	 // glutPostRedisplay();
}

//---------------------------------------------------------
//   Set up the view

void setUpView() {

   // this code initializes the viewing transform
   glLoadIdentity();

   // moves viewer along coordinate axes

   gluLookAt(eye.x, eye.y, eye.z, eyeTarget.x, eyeTarget.y, eyeTarget.z, 0, 1, 0);

   // move the view back some relative to viewer[] position
   glTranslatef(0.0f,0.0f, 0.0f);

   // rotates view
   glRotatef(theta[0], 1.0, 0.0, 0.0);
   glRotatef(theta[1], 0.0, 1.0, 0.0);
   glRotatef(theta[2], 0.0, 0.0, 1.0);

   return;
}


//--------------------------------------------------------
//  Set up the objects
void drawPoint() {
   GLfloat myPos[4] = {0,0,0,1};

   // draw all the particles

      
      glPushMatrix();

    
   cgGLSetStateMatrixParameter(modelViewProj, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);

   for (int i=0; i<NUM_PARTICLES; i++) {
      glBindTexture(GL_TEXTURE_2D, 1);

      cgGLEnableProfile(CG_PROFILE_ARBVP1);   
      cgGLEnableProfile(fragProfile);

    
      glBegin(GL_QUADS);
      {
        
       
         cgGLSetParameter3f(vInitial, vel[i][0], vel[i][1], vel[i][2]);
         cgGLSetParameter1f(tInitial, timeCnt[i]);
         cgGLSetParameter1f(cgMass, mass[i]);
      
         if(on2D){
            glTexCoord2f(0,0);glVertex3f(-.1, -.1,0);
            glTexCoord2f(1,0);glVertex3f(+.1, -.1,0);
            glTexCoord2f(1,1);glVertex3f(+.1, +.1,0);
            glTexCoord2f(0,1);glVertex3f(-.1, +.1,0);
         }
         else{
            if(rand() % 2 ? 1 : 0){
               glTexCoord2f(0,0);glVertex3f(-.1, -.1,0);
               glTexCoord2f(1,0);glVertex3f(+.1, -.1,0);
               glTexCoord2f(1,1);glVertex3f(+.1, +.1,0);
               glTexCoord2f(0,1);glVertex3f(-.1, +.1,0);
            }
            else{
               glTexCoord2f(1,0);glVertex3f(0,+.1, -.1);
               glTexCoord2f(1,1);glVertex3f(0,+.1, +.1);
               glTexCoord2f(0,0);glVertex3f(0,-.1, -.1);
               glTexCoord2f(0,1);glVertex3f(0,-.1, +.1);
            }
         }       
      }
      glEnd();
    
      cgGLDisableProfile(fragProfile);
      cgGLDisableProfile(CG_PROFILE_ARBVP1);    
   }
  
     // recover the transform state
   glPopMatrix(); 

   return;
}


//-----------------------------------------------------------
//  Callback functions

void reshapeCallback(int w, int h) {
   // from Angel, p.562
  GW = w;
  GH = h;
   glViewport(0,0,w,h);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   if (w < h) {
      glFrustum(-2.0, 2.0, -2.0*(GLfloat) h / (GLfloat) w,
                2.0*(GLfloat) h / (GLfloat) w, 2.0, 60.0);
   }
   else {
      glFrustum(-2.0, 2.0, -2.0*(GLfloat) w / (GLfloat) h,
                2.0*(GLfloat) w / (GLfloat) h, 2.0, 60.0);
   }

   glMatrixMode(GL_MODELVIEW);
}


void timeStep(int step) {
   // animation code goes here
   
   count += TIMESTEP;
   cgGLSetParameter1f(globalTime, count);

   for(int i = 0; i < NUM_PARTICLES; i++){
      if(count - timeCnt[i] > WAIT_TIME){
         //cgGLSetParameter1f(tInitial, count);
         myCnt = count;
         initializeParticleDataStructure(i);
         timeCnt[i] = count;         
      }
   }
  


   display();

   glutTimerFunc(50,timeStep, 0);

}

//---------------------------------------------------------
//  Particle system routines

void initializeParticleDataStructure(int i) {
   // initial values for all the particles

   
      mass[i] = MAX_MASS * ((float)(rand() % (X_PREC + 1))  / (float)X_PREC);
   
      pos[i][0] = 0;
      pos[i][1] = 0.0;
      pos[i][2] = 0;

      vel[i][0] = ((rand() % 2) ? 1 : -1) * MAX_X_VEL * ((float)(rand() % (X_PREC + 1))  / (float)X_PREC);
      vel[i][1] = ((float)MAX_Y_VEL  * ((float)(rand() % (Y_PREC + 1))  / (float)Y_PREC) ) + MIN_Y_VEL;
      vel[i][2] = ((rand() % 2) ? 1 : -1) * MAX_Z_VEL * ((float)(rand() % (Z_PREC + 1))  / (float)Z_PREC);

      
      acc[i][0] = 0.0;
      acc[i][1] = 0.0;
      acc[i][2] = 0.0;
      
    
      timeCnt[i] = (float)WAIT_TIME * (float)i / (float)NUM_PARTICLES;
   
}


  
//---------------------------------------------------------
//  Main routines


void display (void) {
   // this code executes whenever the window is redrawn (when opened,
   //   moved, resized, etc.
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   
   cgGLBindProgram(program);

   // set the viewing transform
   setUpView();

   drawPoint();
 
   glutSwapBuffers();
}

void myInit(){
   on2D = 0;
    GW = 500;
   GH = 500;
   count = WAIT_TIME;
   myCnt = 0;
   radius = 10;
   phi = 0;
   eye.x = 0;
  eye.y = 0;
  eye.z = 10;
  eyeTarget.x = 0;
  eyeTarget.y = 0;
  eyeTarget.z = 0;
   mytheta = (float) PI;
  
}
// create a double buffered 500x500 pixel color window
int main(int argc, char* argv[]) {
   myInit();
  
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(GW, GH);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("Nick Cross 570");
  
   glDisable(GL_LIGHTING);
   init_tex();
   LoadTexture("particle3.bmp", 1);

   context = cgCreateContext();
  
   if (context == NULL)
	{
		MessageBox(NULL, "Failed To Create Cg Context", "Error", MB_OK);
		return FALSE;													// We Cannot Continue
	}  

   
   vertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
   cgGLSetOptimalOptions(vertexProfile);	
   program = cgCreateProgramFromFile(context, CG_SOURCE, "particle.cg", CG_PROFILE_ARBVP1, "main", 0);


   cgGLLoadProgram(program);

   pInitial		= cgGetNamedParameter(program, "pInitial");
   vInitial		= cgGetNamedParameter(program, "vInitial");
   tInitial		= cgGetNamedParameter(program, "tInitial");
   cgForce     = cgGetNamedParameter(program, "force"); 
   cgMass     = cgGetNamedParameter(program, "mass"); 
   gravity		= cgGetNamedParameter(program, "gravity");
   globalTime		= cgGetNamedParameter(program, "globalTime");
   modelViewProj = cgGetNamedParameter(program, "modelViewProj");


   cgGLEnableProfile(CG_PROFILE_ARBVP1);    
   
      cgGLSetParameter3f(cgForce, 0, 0, 0);
      cgGLSetParameter1f(cgMass, 1);         
      cgGLSetParameter1f(tInitial, 0);
      cgGLSetParameter1f(globalTime, count);
      cgGLSetParameter3f(gravity, 0, -9.8, 0);
      cgGLSetStateMatrixParameter(modelViewProj, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);

   cgGLDisableProfile(CG_PROFILE_ARBVP1);   

  
   fragProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
   cgGLSetOptimalOptions(fragProfile);

   fragProgram = cgCreateProgramFromFile(
        context,
        CG_SOURCE,
        "particleFrag.cg",
        fragProfile,
        "frag_particle",    // entry point
        0);   // arguments

    cgGLLoadProgram(fragProgram);

    cgGLBindProgram(fragProgram);
    cgGLEnableProfile(fragProfile);
    cgGLDisableProfile(fragProfile);



   for (int i=0; i<NUM_PARTICLES; i++) {
      initializeParticleDataStructure(i);
   }

   // this sets the size of a point in pixels
   glPointSize(5.0);

	glEnable(GL_DEPTH_TEST);
	glutDisplayFunc(display);
   glutKeyboardFunc(keyCallback);
   glutMouseFunc(mouse);
  glutMotionFunc(mouseMove);
   glutReshapeFunc(reshapeCallback);
	glutTimerFunc(50,timeStep, 0);  // 50 millisecond callback
	glutMainLoop();
	return 0;
}

//---------------------------------------------------------
//  Utility functions

void normalize(float v[3]) {
   // normalize v[] and return the result in v[]
   // from OpenGL Programming Guide, p. 58
   GLfloat d = sqrt(v[1]*v[1] + v[2]*v[2] + v[3]*v[3]);
   if (d == 0.0) {
      printf("zero length vector");
      return;
   }
   v[1] = v[1]/d; v[2] = v[2]/d; v[3] = v[3]/d;
}

void normCrossProd(float v1[3], float v2[3], float out[3]) {
   // cross v1[] and v2[] and return the result in out[]
   // from OpenGL Programming Guide, p. 58
   out[0] = v1[1]*v2[2] - v1[2]*v2[1];
   out[1] = v1[2]*v2[0] - v2[0]*v2[2];
   out[2] = v1[0]*v2[1] - v1[1]*v2[0];
   normalize(out);
}

